@@ -1,15 +1,21 @@
 import sys
 from contextlib import contextmanager
+from datetime import datetime
 from enum import Enum, auto
 
 import click
 import logging
 import getpass
 
+import tzlocal
+from asn1crypto import pem, cms
+
+from pyhanko.pdf_utils.layout import LayoutError
 from pyhanko_certvalidator import ValidationContext
 from pyhanko.config import (
     init_validation_context_kwargs, parse_cli_config,
-    CLIConfig, LogConfig, StdLogOutput, parse_logging_config
+    CLIConfig, LogConfig, StdLogOutput, parse_logging_config,
+    PKCS11SignatureConfig
 )
 from pyhanko.pdf_utils import misc
 from pyhanko.pdf_utils.config_utils import ConfigurationError
@@ -78,6 +84,9 @@ def pyhanko_exception_manager():
     except SigningError as e:
         exception = e
         msg = "Error raised while producing signed file."
+    except LayoutError as e:
+        exception = e
+        msg = "Error raised while producing signature layout."
     except Exception as e:
         exception = e
         msg = "Generic processing error."
@@ -99,6 +108,7 @@ class Ctx(Enum):
     STAMP_URL = auto()
     NEW_FIELD_SPEC = auto()
     PREFER_PSS = auto()
+    DETACH_PEM = auto()
 
 
 @click.group()
@@ -165,7 +175,7 @@ def cli(ctx, config, verbose):
         logging.debug('There was no configuration to parse.')
 
 
-@cli.group(help='sign PDF files', name='sign')
+@cli.group(help='sign PDFs and other files', name='sign')
 def signing():
     pass
 
@@ -306,9 +316,8 @@ def _select_style(ctx, style_name, url):
     return style
 
 
-def _signature_status(ltv_profile, force_revinfo, soft_revocation_check,
-                      pretty_print, vc_kwargs, key_usage_settings,
-                      executive_summary, embedded_sig):
+def _prepare_vc(vc_kwargs, soft_revocation_check, force_revinfo):
+
     if soft_revocation_check and force_revinfo:
         raise click.ClickException(
             "--soft-revocation-check is incompatible with "
@@ -321,20 +330,49 @@ def _signature_status(ltv_profile, force_revinfo, soft_revocation_check,
     else:
         rev_mode = 'hard-fail'
     vc_kwargs['revocation_mode'] = rev_mode
+    return vc_kwargs
+
+
+def _signature_status(ltv_profile, vc_kwargs, force_revinfo, key_usage_settings,
+                      embedded_sig):
+    if ltv_profile is None:
+        vc = ValidationContext(**vc_kwargs)
+        status = validation.validate_pdf_signature(
+            embedded_sig, key_usage_settings=key_usage_settings,
+            signer_validation_context=vc
+        )
+    else:
+        status = validation.validate_pdf_ltv_signature(
+            embedded_sig, ltv_profile,
+            key_usage_settings=key_usage_settings,
+            force_revinfo=force_revinfo,
+            validation_context_kwargs=vc_kwargs
+        )
+    return status
+
+
+def _validate_detached(infile, sig_infile, validation_context,
+                       key_usage_settings):
+    sig_bytes = sig_infile.read()
     try:
-        if ltv_profile is None:
-            vc = ValidationContext(**vc_kwargs)
-            status = validation.validate_pdf_signature(
-                embedded_sig, key_usage_settings=key_usage_settings,
-                signer_validation_context=vc
-            )
-        else:
-            status = validation.validate_pdf_ltv_signature(
-                embedded_sig, ltv_profile,
-                key_usage_settings=key_usage_settings,
-                force_revinfo=force_revinfo,
-                validation_context_kwargs=vc_kwargs
-            )
+        if pem.detect(sig_bytes):
+            _, _, sig_bytes = pem.unarmor(sig_bytes)
+        content_info = cms.ContentInfo.load(sig_bytes)
+        if content_info['content_type'].native != 'signed_data':
+            raise click.ClickException("CMS content type is not signedData")
+    except ValueError as e:
+        raise click.ClickException("Could not parse CMS object") from e
+
+    return validation.validate_detached_cms(
+        infile, signed_data=content_info['content'],
+        signer_validation_context=validation_context,
+        key_usage_settings=key_usage_settings
+    )
+
+
+def _signature_status_str(status_callback, pretty_print, executive_summary):
+    try:
+        status = status_callback()
         if executive_summary and not pretty_print:
             return 'VALID' if status.bottom_line else 'INVALID'
         elif pretty_print:
@@ -392,12 +430,19 @@ def _signature_status(ltv_profile, force_revinfo, soft_revocation_check,
               type=bool, is_flag=True, default=False, show_default=True)
 @click.option('--password', required=False, type=str,
               help='password to access the file (can also be read from stdin)')
+@click.option(
+    '--detached', type=click.File('rb'),
+    help=(
+        'Read signature CMS object from the indicated file; '
+        'this can be used to verify signatures on non-PDF files'
+    )
+)
 @click.pass_context
 def validate_signatures(ctx, infile, executive_summary,
                         pretty_print, validation_context, trust, trust_replace,
                         other_certs, ltv_profile, force_revinfo,
                         soft_revocation_check, no_revocation_check, password,
-                        retroactive_revinfo):
+                        retroactive_revinfo, detached):
 
     if no_revocation_check:
         soft_revocation_check = True
@@ -417,7 +462,22 @@ def validate_signatures(ctx, infile, executive_summary,
     )
 
     key_usage_settings = _get_key_usage_settings(ctx, validation_context)
+    vc_kwargs = _prepare_vc(
+        vc_kwargs, soft_revocation_check=soft_revocation_check,
+        force_revinfo=force_revinfo
+    )
     with pyhanko_exception_manager():
+        if detached is not None:
+            status_str = _signature_status_str(
+                status_callback=lambda: _validate_detached(
+                    infile, detached, ValidationContext(**vc_kwargs),
+                    key_usage_settings
+                ),
+                pretty_print=pretty_print, executive_summary=executive_summary
+            )
+            print(status_str)
+            return
+
         r = PdfFileReader(infile)
         sh = r.security_handler
         if isinstance(sh, crypt.StandardSecurityHandler):
@@ -434,10 +494,13 @@ def validate_signatures(ctx, infile, executive_summary,
 
         for ix, embedded_sig in enumerate(r.embedded_regular_signatures):
             fingerprint: str = embedded_sig.signer_cert.sha256.hex()
-            status_str = _signature_status(
-                ltv_profile, force_revinfo, soft_revocation_check,
-                pretty_print, vc_kwargs, key_usage_settings,
-                executive_summary, embedded_sig
+            status_str = _signature_status_str(
+                status_callback=lambda: _signature_status(
+                    ltv_profile=ltv_profile, force_revinfo=force_revinfo,
+                    vc_kwargs=vc_kwargs, key_usage_settings=key_usage_settings,
+                    embedded_sig=embedded_sig
+                ),
+                pretty_print=pretty_print, executive_summary=executive_summary
             )
             name = embedded_sig.field_name
 
@@ -572,6 +635,17 @@ def ltv_fix(ctx, infile, field, timestamp_url, apply_lta_timestamp,
     required=False, type=str
 )
 @trust_options
+@click.option(
+    '--detach', type=bool, is_flag=True, default=False,
+    help=(
+        'write only the signature CMS object to the output file; '
+        'this can be used to sign non-PDF files'
+    )
+)
+@click.option(
+    '--detach-pem', help='output PEM data instead of DER when using --detach',
+    type=bool, is_flag=True, default=False
+)
 @click.option('--retroactive-revinfo',
               help='Treat revocation info as retroactively valid '
                    '(i.e. ignore thisUpdate timestamp)',
@@ -580,11 +654,17 @@ def ltv_fix(ctx, infile, field, timestamp_url, apply_lta_timestamp,
 def addsig(ctx, field, name, reason, location, certify, existing_only,
            timestamp_url, use_pades, with_validation_info,
            validation_context, trust_replace, trust, other_certs,
-           style_name, stamp_url, prefer_pss, retroactive_revinfo):
+           style_name, stamp_url, prefer_pss, retroactive_revinfo,
+           detach, detach_pem):
     ctx.obj[Ctx.EXISTING_ONLY] = existing_only or field is None
     ctx.obj[Ctx.TIMESTAMP_URL] = timestamp_url
     ctx.obj[Ctx.PREFER_PSS] = prefer_pss
 
+    if detach:
+        ctx.obj[Ctx.DETACH_PEM] = detach_pem
+        ctx.obj[Ctx.SIG_META] = None
+        return  # everything else doesn't apply
+
     if use_pades:
         subfilter = fields.SigSeedSubFilter.PADES
     else:
@@ -659,6 +739,32 @@ def get_text_params(ctx):
     return text_params
 
 
+def detached_sig(signer: signers.Signer, infile_path, outfile,
+                 timestamp_url, use_pem):
+
+    with pyhanko_exception_manager():
+        if timestamp_url is not None:
+            timestamper = HTTPTimeStamper(timestamp_url)
+            timestamp = None
+        else:
+            timestamper = None
+            # in this case, embed the signing time as an unsigned attr
+            timestamp = datetime.now(tz=tzlocal.get_localzone())
+
+        with open(infile_path, 'rb') as inf:
+            signature = signer.sign_general_data(
+                inf, signers.DEFAULT_MD, timestamper=timestamper,
+                timestamp=timestamp
+            )
+
+        output_bytes = signature.dump()
+        if use_pem:
+            output_bytes = pem.armor('PKCS7', output_bytes)
+
+        # outfile is managed by Click
+        outfile.write(output_bytes)
+
+
 def addsig_simple_signer(signer: signers.SimpleSigner, infile_path, outfile,
                          timestamp_url, signature_meta, existing_fields_only,
                          style, text_params, new_field_spec):
@@ -672,7 +778,7 @@ def addsig_simple_signer(signer: signers.SimpleSigner, infile_path, outfile,
             signer_key=signer.signing_key
         )
 
-        generic_sign(
+        generic_sign_pdf(
             writer=writer, outfile=outfile,
             signature_meta=signature_meta, signer=signer,
             timestamper=timestamper, style=style, new_field_spec=new_field_spec,
@@ -680,8 +786,8 @@ def addsig_simple_signer(signer: signers.SimpleSigner, infile_path, outfile,
         )
 
 
-def generic_sign(*, writer, outfile, signature_meta, signer, timestamper,
-                 style, new_field_spec, existing_fields_only, text_params):
+def generic_sign_pdf(*, writer, outfile, signature_meta, signer, timestamper,
+                     style, new_field_spec, existing_fields_only, text_params):
     result = signers.PdfSigner(
         signature_meta, signer=signer, timestamper=timestamper,
         stamp_style=style, new_field_spec=new_field_spec
@@ -730,7 +836,13 @@ def addsig_pemder(ctx, infile, outfile, key, cert, chain, passfile):
         cert_file=cert, key_file=key, key_passphrase=passphrase,
         ca_chain_files=chain, prefer_pss=ctx.obj[Ctx.PREFER_PSS]
     )
-    return addsig_simple_signer(
+
+    if ctx.obj[Ctx.SIG_META] is None:
+        detached_sig(
+            signer, infile, outfile, timestamp_url=timestamp_url,
+            use_pem=ctx.obj[Ctx.DETACH_PEM]
+        )
+    addsig_simple_signer(
         signer, infile, outfile, timestamp_url=timestamp_url,
         signature_meta=signature_meta,
         existing_fields_only=existing_fields_only,
@@ -771,7 +883,12 @@ def addsig_pkcs12(ctx, infile, outfile, pfx, chain, passfile):
         pfx_file=pfx, passphrase=passphrase, ca_chain_files=chain,
         prefer_pss=ctx.obj[Ctx.PREFER_PSS]
     )
-    return addsig_simple_signer(
+    if ctx.obj[Ctx.SIG_META] is None:
+        detached_sig(
+            signer, infile, outfile, timestamp_url=timestamp_url,
+            use_pem=ctx.obj[Ctx.DETACH_PEM]
+        )
+    addsig_simple_signer(
         signer, infile, outfile, timestamp_url=timestamp_url,
         signature_meta=signature_meta,
         existing_fields_only=existing_fields_only,
@@ -780,59 +897,83 @@ def addsig_pkcs12(ctx, infile, outfile, pfx, chain, passfile):
     )
 
 
-# TODO add options to specify extra certs to include
+def _sign_pkcs11(ctx, signer, infile, outfile, timestamp_url):
+    with pyhanko_exception_manager():
+        if ctx.obj[Ctx.SIG_META] is None:
+            return detached_sig(
+                signer, infile, outfile, timestamp_url=timestamp_url,
+                use_pem=ctx.obj[Ctx.DETACH_PEM]
+            )
 
-@click.argument('infile', type=click.File('rb'))
+        if timestamp_url is not None:
+            timestamper = HTTPTimeStamper(timestamp_url)
+        else:
+            timestamper = None
+
+        with open(infile, 'rb') as inf:
+            generic_sign_pdf(
+                writer=IncrementalPdfFileWriter(inf),
+                outfile=outfile,
+                signature_meta=ctx.obj[Ctx.SIG_META], signer=signer,
+                timestamper=timestamper, style=ctx.obj[Ctx.STAMP_STYLE],
+                new_field_spec=ctx.obj[Ctx.NEW_FIELD_SPEC],
+                existing_fields_only=ctx.obj[Ctx.EXISTING_ONLY],
+                text_params=get_text_params(ctx)
+            )
+
+
+@click.argument('infile', type=readable_file)
 @click.argument('outfile', type=click.File('wb'))
 @click.option('--lib', help='path to PKCS#11 module',
-              type=readable_file, required=True)
+              type=readable_file, required=False)
 @click.option('--token-label', help='PKCS#11 token label', type=str,
-              required=True)
-@click.option('--cert-label', help='certificate label', type=str, required=True)
+              required=False)
+@click.option('--cert-label', help='certificate label', type=str,
+              required=False)
 @click.option('--key-label', help='key label', type=str, required=False)
 @click.option('--slot-no', help='specify PKCS#11 slot to use',
               required=False, type=int, default=None)
 @click.option('--skip-user-pin', type=bool, show_default=True,
               default=False, required=False, is_flag=True,
               help='do not prompt for PIN (e.g. if the token has a PIN pad)')
+@click.option('--p11-setup', type=str, required=False,
+              help='name of preconfigured PKCS#11 profile (overrides all '
+                   'other options)')
 @click.pass_context
 def addsig_pkcs11(ctx, infile, outfile, lib, token_label,
-                  cert_label, key_label, slot_no, skip_user_pin):
+                  cert_label, key_label, slot_no, skip_user_pin, p11_setup):
     from pyhanko.sign import pkcs11
-    signature_meta = ctx.obj[Ctx.SIG_META]
-    existing_fields_only = ctx.obj[Ctx.EXISTING_ONLY]
     timestamp_url = ctx.obj[Ctx.TIMESTAMP_URL]
 
-    if skip_user_pin:
-        user_pin = None
+    if p11_setup:
+        cli_config: CLIConfig = ctx.obj.get(Ctx.CLI_CONFIG, None)
+        if cli_config is None:
+            raise click.ClickException(
+                "The --p11-setup option requires a configuration file"
+            )
+        try:
+            pkcs11_config = cli_config.get_pcks11_config(p11_setup)
+        except ConfigurationError as e:
+            msg = f"Error while reading PKCS#11 config {p11_setup}"
+            logger.error(msg, exc_info=e)
+            raise click.ClickException(msg)
     else:
-        user_pin = getpass.getpass(prompt='PKCS#11 user PIN: ')
-
-    session = pkcs11.open_pkcs11_session(
-        lib_location=lib, slot_no=slot_no, token_label=token_label,
-        user_pin=user_pin
-    )
-    if timestamp_url is not None:
-        timestamper = HTTPTimeStamper(timestamp_url)
-    else:
-        timestamper = None
-
-    signer = pkcs11.PKCS11Signer(
-        session, cert_label=cert_label, key_label=key_label,
-    )
-
-    with pyhanko_exception_manager():
-        generic_sign(
-            writer=IncrementalPdfFileWriter(infile), outfile=outfile,
-            signature_meta=signature_meta, signer=signer,
-            timestamper=timestamper, style=ctx.obj[Ctx.STAMP_STYLE],
-            new_field_spec=ctx.obj[Ctx.NEW_FIELD_SPEC],
-            existing_fields_only=existing_fields_only,
-            text_params=get_text_params(ctx)
+        if not (lib and token_label and cert_label):
+            raise click.ClickException(
+                "The parameters --lib, --token-label and --cert-label "
+                "are required."
+            )
+        pkcs11_config = PKCS11SignatureConfig(
+            module_path=lib, cert_label=cert_label, key_label=key_label,
+            slot_no=slot_no, token_label=token_label,
+            prompt_pin=not skip_user_pin
         )
 
+    with pkcs11.PKCS11SigningContext(pkcs11_config) as signer:
+        _sign_pkcs11(ctx, signer, infile, outfile, timestamp_url)
 
-@click.argument('infile', type=click.File('rb'))
+
+@click.argument('infile', type=readable_file)
 @click.argument('outfile', type=click.File('wb'))
 @click.option('--lib', help='path to libbeidpkcs11 library file',
               type=readable_file, required=True)
@@ -844,26 +985,11 @@ def addsig_pkcs11(ctx, infile, outfile, lib, token_label,
 @click.pass_context
 def addsig_beid(ctx, infile, outfile, lib, use_auth_cert, slot_no):
     from pyhanko.sign import beid
-    signature_meta = ctx.obj[Ctx.SIG_META]
-    existing_fields_only = ctx.obj[Ctx.EXISTING_ONLY]
     timestamp_url = ctx.obj[Ctx.TIMESTAMP_URL]
     session = beid.open_beid_session(lib, slot_no=slot_no)
-    if timestamp_url is not None:
-        timestamper = HTTPTimeStamper(timestamp_url)
-    else:
-        timestamper = None
-
-    signer = beid.BEIDSigner(session, use_auth_cert=use_auth_cert)
-
-    with pyhanko_exception_manager():
-        generic_sign(
-            writer=IncrementalPdfFileWriter(infile), outfile=outfile,
-            signature_meta=signature_meta, signer=signer,
-            timestamper=timestamper, style=ctx.obj[Ctx.STAMP_STYLE],
-            new_field_spec=ctx.obj[Ctx.NEW_FIELD_SPEC],
-            existing_fields_only=existing_fields_only,
-            text_params=get_text_params(ctx)
-        )
+    with session:
+        signer = beid.BEIDSigner(session, use_auth_cert=use_auth_cert)
+        _sign_pkcs11(ctx, signer, infile, outfile, timestamp_url)
 
 
 def _pkcs11_cmd(name, hlp, fun):
