@@ -7,12 +7,14 @@ for the original license.
 
 import os
 import struct
+import enum
+from dataclasses import dataclass
 from io import BytesIO
-from typing import List, Union, Optional, Tuple
+from typing import List, Union, Optional, Tuple, Iterable
 
 from asn1crypto import x509
 
-from pyhanko.pdf_utils import generic
+from pyhanko.pdf_utils import generic, content
 from pyhanko.pdf_utils.crypt import (
     SecurityHandler, StandardSecurityHandler,
     PubKeySecurityHandler,
@@ -29,7 +31,7 @@ from pyhanko import __version__
 __all__ = [
     'ObjectStream', 'BasePdfFileWriter',
     'PageObject', 'PdfFileWriter', 'init_xobject_dictionary',
-    'copy_into_new_writer'
+    'copy_into_new_writer', 'DeveloperExtension', 'DevExtensionMultivalued'
 ]
 
 VENDOR = 'pyHanko ' + __version__
@@ -113,6 +115,132 @@ class ObjectStream:
         return stream_object
 
 
+class DevExtensionMultivalued(enum.Enum):
+    """
+    Setting indicating how an extension is expected to behave well w.r.t.
+    the new mechanism for multivalued extensions in ISO 32000-2:2020.
+    """
+
+    ALWAYS = enum.auto()
+    """
+    Always serialise this extension as a multivalued extension.
+    """
+
+    NEVER = enum.auto()
+    """
+    Never serialise this extension as a multivalued extension.
+    """
+
+    MAYBE = enum.auto()
+    """
+    Make this extension single-valued whenever possible, but allow multiple
+    values as well, e.g. when a different but non-comparable extension with
+    the same prefix is already present in the file.
+    """
+
+
+@dataclass(frozen=True)
+class DeveloperExtension:
+    """
+    PDF developer extension designation.
+    """
+
+    prefix_name: generic.NameObject
+    """
+    Registered developer prefix.
+    """
+
+    base_version: generic.NameObject
+    """
+    Base version on to which the extension applies.
+    """
+
+    extension_level: int
+    """
+    Extension level.
+    """
+
+    url: Optional[str] = None
+    """
+    Optional URL linking to the extension's documentation.
+    """
+
+    extension_revision: Optional[str] = None
+    """
+    Optional extra revision information. Not comparable.
+    """
+
+    compare_by_level: bool = False
+    """
+    Compare developer extensions by level number.
+    If this value is ``True`` and a copy of this extension already exists in the
+    target file with a higher level number, do not override it.
+    If one exists with a lower level number, override it.
+    
+    If this value is ``False``, the decision is based on :attr:`subsumed_by`
+    and :attr:`subsumes`.
+    
+    .. warning::
+        It is generally not safe to assume that extension levels are used as a
+        versioning system (i.e. that higher extension levels supersede lower
+        ones), hence why the default is ``False``.
+    """
+
+    subsumed_by: Iterable[int] = ()
+    """
+    List of extension levels that would subsume this one. If one of these is
+    present in the extensions dictionary, attempting to register this extension
+    will not override it.
+    
+    Default value: empty.
+    
+    .. warning::
+        This parameter is ignored if :class:`compare_by_level` is ``True``.
+    """
+
+    subsumes: Iterable[int] = ()
+    """
+    List of extensions explicitly subsumed by this one. If one of these is
+    present in the extensions dictionary, attempting to register this extension
+    will override it.
+    
+    Default value: empty.
+    
+    .. warning::
+        This parameter is ignored if :class:`compare_by_level` is ``True``.
+    """
+
+    multivalued: DevExtensionMultivalued = DevExtensionMultivalued.MAYBE
+    """
+    Setting indicating whether this extension is expected to behave well w.r.t.
+    the new mechanism for multivalued extensions in ISO 32000-2:2020.
+    """
+
+    def as_pdf_object(self) -> generic.DictionaryObject:
+        """
+        Format the data in this object into a PDF dictionary for registration
+        into the `/Extensions` dictionary.
+
+        :return:
+            A :class:`.generic.DictionaryObject`.
+        """
+
+        result = generic.DictionaryObject({
+            pdf_name('/Type'): pdf_name('/DeveloperExtensions'),
+            pdf_name('/BaseVersion'): self.base_version,
+            pdf_name('/ExtensionLevel'): generic.NumberObject(
+                self.extension_level
+            ),
+        })
+        if self.url is not None:
+            result['/URL'] = generic.TextStringObject(self.url)
+        if self.extension_revision is not None:
+            result['/ExtensionRevision'] = generic.TextStringObject(
+                self.extension_revision
+            )
+        return result
+
+
 def _contiguous_xref_chunks(position_dict):
     """
     Helper method to divide the XRef table (or stream) into contiguous chunks.
@@ -303,6 +431,22 @@ class BasePdfFileWriter(PdfHandler):
         else:
             self._info = info
 
+        self._font_resources = {}
+
+    def get_subset_collection(self, base_postscript_name: str):
+        from .font.api import FontSubsetCollection
+        try:
+            fsc = self._font_resources[base_postscript_name]
+        except KeyError:
+            fsc = FontSubsetCollection(base_postscript_name)
+            self._font_resources[base_postscript_name] = fsc
+
+        return fsc
+
+    def ensure_output_version(self, version):
+        if self.output_version < version:
+            self.output_version = version
+
     def set_info(self, info: Optional[Union[generic.IndirectObject,
                                       generic.DictionaryObject]]):
         """
@@ -365,6 +509,105 @@ class BasePdfFileWriter(PdfHandler):
         """
         pass
 
+    def register_extension(self, ext: DeveloperExtension):
+        try:
+            extensions = self.root['/Extensions']
+        except KeyError:
+            self.root['/Extensions'] = extensions = generic.DictionaryObject()
+
+        # check if the extension is already registered,
+        cur_ext_value = extensions.get(ext.prefix_name, None)
+        extension_dicts = ()
+        old_ext_multivalued = False
+        if isinstance(cur_ext_value, generic.DictionaryObject):
+            extension_dicts = cur_ext_value,
+        elif isinstance(cur_ext_value, generic.ArrayObject):
+            extension_dicts = tuple(cur_ext_value)
+            old_ext_multivalued = True
+        elif cur_ext_value is not None:
+            cls_name = type(cur_ext_value).__name__
+            raise PdfReadError(
+                f"PDF extension value is of type {cls_name}, "
+                f"expected (direct) PDF dictionary or array."
+            )
+
+        # investigate the relationship between the new extension's extension
+        # level and that of any existing extension dictionaries with the same
+        # prefix
+        for ix, ext_dict in enumerate(extension_dicts):
+            if not isinstance(ext_dict, generic.DictionaryObject):
+                cls_name = type(ext_dict).__name__
+                raise PdfReadError(
+                    f"PDF extension array entry is of type "
+                    f"{cls_name}, expected (direct) PDF dictionary"
+                )
+            try:
+                lvl = int(ext_dict.raw_get('/ExtensionLevel'))
+            except (TypeError, ValueError, KeyError):
+                raise PdfReadError(
+                    "Could not read developer extension level"
+                )
+            # TODO is this still appropriate with the new ExtensionRevision
+            #  values?
+            if lvl == ext.extension_level:
+                old_ext_applies = True
+                replace_old = False
+            elif ext.compare_by_level:
+                old_ext_applies = lvl >= ext.extension_level
+                replace_old = not old_ext_applies
+            else:
+                # this covers the case where there's no obvious comparison
+                old_ext_applies = lvl in ext.subsumed_by
+                replace_old = lvl in ext.subsumes
+            if old_ext_applies:
+                return  # nothing to do, old extension doesn't need replacing
+            elif replace_old:
+                # New extension takes priority,
+                # replace the old extension
+                if old_ext_multivalued:
+                    # it was an array, so replace the value at the proper
+                    # index
+                    cur_ext_value[ix] = ext.as_pdf_object()
+                elif ext.multivalued == DevExtensionMultivalued.ALWAYS:
+                    # extension was previously not multivalued, but we now
+                    # make it so
+                    extensions[ext.prefix_name] = generic.ArrayObject(
+                        [ext.as_pdf_object()]
+                    )
+                else:
+                    extensions[ext.prefix_name] = ext.as_pdf_object()
+                self.update_container(extensions)
+                return  # we're done here
+            elif ext.multivalued == DevExtensionMultivalued.NEVER:
+                # there is a matching extension in the file, but we can't
+                # replace it -> error
+                raise PdfWriteError(
+                    f"Could not register extension with prefix "
+                    f"{ext.prefix_name} and level {ext.extension_level}; "
+                    f"file contains extension with same prefix and "
+                    f"extension level {lvl}. If this extension level is "
+                    f"safe to override, mark it as subsumed."
+                )
+        # if we got here, we're going to add the new extension without
+        # replacing any existing ones
+        ext_dict = ext.as_pdf_object()
+        if cur_ext_value is None:
+            # nothing there yet, just write the extension
+            if ext.multivalued == DevExtensionMultivalued.ALWAYS:
+                extensions[ext.prefix_name] = generic.ArrayObject([ext_dict])
+            else:
+                extensions[ext.prefix_name] = ext_dict
+        else:
+            if old_ext_multivalued:
+                # there was already an array, so just tack on our new value
+                cur_ext_value.append(ext_dict)
+            else:
+                # turn the extension dict into an array
+                extensions[ext.prefix_name] = generic.ArrayObject(
+                    [cur_ext_value, ext_dict]
+                )
+        self.update_container(extensions)
+
     def get_object(self, ido):
         if ido.pdf not in self._resolves_objs_from:
             raise ValueError(
@@ -527,6 +770,11 @@ class BasePdfFileWriter(PdfHandler):
 
     def _write(self, stream, skip_header=False):
 
+        # ensure that all font resources are flushed
+        for fsc in self._font_resources.values():
+            for engine in fsc.subsets.values():
+                engine.prepare_write()
+
         object_positions = {}
 
         if self.stream_xrefs:
@@ -727,7 +975,7 @@ class BasePdfFileWriter(PdfHandler):
             return obj
 
     def import_page_as_xobject(self, other: PdfHandler, page_ix=0,
-                               content_stream=0, inherit_filters=True):
+                               inherit_filters=True):
         """
         Import a page content stream from some other
         :class:`~.rw_common.PdfHandler` into the current one as a form XObject.
@@ -736,9 +984,6 @@ class BasePdfFileWriter(PdfHandler):
             A :class:`~.rw_common.PdfHandler`
         :param page_ix:
             Index of the page to copy (default: 0)
-        :param content_stream:
-            Index of the page's content stream to copy, if multiple are present
-            (default: 0)
         :param inherit_filters:
             Inherit the content stream's filters, if present.
         :return:
@@ -772,7 +1017,21 @@ class BasePdfFileWriter(PdfHandler):
         # if the page /Contents is an array, retrieve the content stream
         # with the appropriate index
         if isinstance(command_stream, generic.ArrayObject):
-            command_stream = command_stream[content_stream].get_object()
+            if len(command_stream) == 1 and inherit_filters:
+                command_stream = command_stream[0].get_object()
+            else:
+                # There are multiple streams, so inheriting filters is not
+                # a well-defined operations.
+                # Decode and concatenate, then put in a flate filter, and return
+                result = generic.StreamObject(
+                    stream_dict,
+                    stream_data=b''.join(
+                        partial_stream.get_object().data
+                        for partial_stream in command_stream
+                    ),
+                )
+                result.compress()
+                return self.add_object(result)
         assert isinstance(command_stream, generic.StreamObject)
         filters = None
         if inherit_filters:
@@ -794,6 +1053,156 @@ class BasePdfFileWriter(PdfHandler):
 
         return self.add_object(result)
 
+    # TODO these can be simplified considerably using the new update_container
+    def add_stream_to_page(self, page_ix, stream_ref, resources=None,
+                           prepend=False):
+        """Append an indirect stream object to a page in a PDF as a content
+        stream.
+
+        :param page_ix:
+            Index of the page to modify.
+            The first page has index `0`.
+        :param stream_ref:
+            :class:`~.generic.IndirectObject` reference to the stream
+            object to add.
+        :param resources:
+            Resource dictionary containing resources to add to the page's
+            existing resource dictionary.
+        :param prepend:
+            Prepend the content stream to the list of content streams, as
+            opposed to appending it to the end.
+            This has the effect of causing the stream to be rendered
+            underneath the already existing content on the page.
+        :return:
+            An :class:`~.generic.IndirectObject` reference to the page object
+            that was modified.
+        """
+
+        page_obj_ref, res_ref = self.find_page_for_modification(page_ix)
+
+        page_obj = page_obj_ref.get_object()
+
+        contents_ref = page_obj.raw_get('/Contents')
+
+        if isinstance(contents_ref, generic.IndirectObject):
+            contents = contents_ref.get_object()
+            if isinstance(contents, generic.ArrayObject):
+                # This is the easy case. It suffices to mark
+                # this array for an update, and append our stream to it.
+                self.mark_update(contents_ref)
+                if prepend:
+                    contents.insert(0, stream_ref)
+                else:
+                    contents.append(stream_ref)
+            elif isinstance(contents, generic.StreamObject):
+                # replace the old stream with an array containing
+                # a reference to the original stream, and our own stream.
+                new = [stream_ref, contents_ref] \
+                    if prepend else [contents_ref, stream_ref]
+                contents = generic.ArrayObject(new)
+                page_obj[pdf_name('/Contents')] = self.add_object(contents)
+                # mark the page to be updated as well
+                self.mark_update(page_obj_ref)
+            else:
+                raise ValueError('Unexpected type for page /Contents')
+        elif isinstance(contents_ref, generic.ArrayObject):
+            # make /Contents an indirect array, and append our stream
+            contents = contents_ref
+            if prepend:
+                contents.insert(0, stream_ref)
+            else:
+                contents.append(stream_ref)
+            page_obj[pdf_name('/Contents')] = self.add_object(contents)
+            self.mark_update(page_obj_ref)
+        else:
+            raise ValueError('Unexpected type for page /Contents')
+
+        if resources is None:
+            return
+
+        if isinstance(res_ref, generic.IndirectObject):
+            # we can get away with only updating this reference
+            orig_resource_dict = res_ref.get_object()
+            assert isinstance(orig_resource_dict, generic.DictionaryObject)
+            if self.merge_resources(orig_resource_dict, resources):
+                self.mark_update(res_ref)
+        else:
+            # don't bother trying to update the resource object, just
+            # clone it and add it to the current page object.
+            orig_resource_dict = generic.DictionaryObject(res_ref)
+            page_obj[pdf_name('/Resources')] = self.add_object(
+                orig_resource_dict
+            )
+            self.merge_resources(orig_resource_dict, resources)
+
+        return page_obj_ref
+
+    def add_content_to_page(self, page_ix, pdf_content: content.PdfContent,
+                            prepend=False):
+        """
+        Convenience wrapper around :meth:`add_stream_to_page` to turn a
+        :class:`~.content.PdfContent` instance into a page content stream.
+
+        :param page_ix:
+            Index of the page to modify.
+            The first page has index `0`.
+        :param pdf_content:
+            An instance of :class:`~.content.PdfContent`
+        :param prepend:
+            Prepend the content stream to the list of content streams, as
+            opposed to appending it to the end.
+            This has the effect of causing the stream to be rendered
+            underneath the already existing content on the page.
+        :return:
+            An :class:`~.generic.IndirectObject` reference to the page object
+            that was modified.
+        """
+        as_stream = generic.StreamObject({}, stream_data=pdf_content.render())
+        return self.add_stream_to_page(
+            page_ix, self.add_object(as_stream),
+            resources=pdf_content.resources.as_pdf_object(), prepend=prepend
+        )
+
+    # TODO this doesn't really belong here
+    def merge_resources(self, orig_dict, new_dict) -> bool:
+        """
+        Update an existing resource dictionary object with data from another
+        one. Returns ``True`` if the original dict object was modified directly.
+
+        The caller is responsible for avoiding name conflicts with existing
+        resources.
+        """
+
+        update_needed = False
+        for key, value in new_dict.items():
+            try:
+                orig_value_ref = orig_dict.raw_get(key)
+            except KeyError:
+                update_needed = True
+                orig_dict[key] = value
+                continue
+
+            if isinstance(orig_value_ref, generic.IndirectObject):
+                orig_value = orig_value_ref.get_object()
+                self.mark_update(orig_value_ref)
+            else:
+                orig_value = orig_value_ref
+                update_needed = True
+
+            if isinstance(orig_value, generic.ArrayObject):
+                # the /ProcSet case
+                orig_value.extend(value)
+            elif isinstance(orig_value, generic.DictionaryObject):
+                for key_, value_ in value.items():
+                    if key_ in orig_value:
+                        raise ValueError(
+                            'Naming conflict in resource of type %s: '
+                            'key %s occurs in both.' % (key, key_)
+                        )
+                    orig_value[key_] = value_
+
+        return update_needed
+
 
 class PageObject(generic.DictionaryObject):
     """Subclass of :class:`~.generic.DictionaryObject` that handles some of the
@@ -822,7 +1231,7 @@ class PageObject(generic.DictionaryObject):
         super().__init__({
             pdf_name('/Type'): pdf_name('/Page'),
             pdf_name('/MediaBox'): generic.ArrayObject(
-                map(generic.NumberObject, media_box)
+                map(generic.FloatObject, media_box)
             ),
             pdf_name('/Resources'): resources,
             pdf_name('/Contents'): contents
