@@ -36,13 +36,15 @@ from cryptography.hazmat.primitives.asymmetric import padding
 
 __all__ = [
     'SignatureStatus', 'simple_cms_attribute', 'find_cms_attribute',
+    'extract_message_digest', 'validate_sig_integrity',
     'CertificateStore', 'SimpleCertificateStore',
     'KeyUsageConstraints',
     'SigningError', 'UnacceptableSignerError', 'WeakHashAlgorithmError',
     'SignatureValidationError',
     'load_certs_from_pemder', 'load_cert_from_pemder',
     'load_private_key_from_pemder', 'get_pyca_cryptography_hash',
-    'DEFAULT_WEAK_HASH_ALGORITHMS'
+    'DEFAULT_WEAK_HASH_ALGORITHMS',
+    'optimal_pss_params', 'as_signing_certificate'
 ]
 
 logger = logging.getLogger(__name__)
@@ -370,6 +372,14 @@ class SignatureStatus:
             logger.warning(f"Chain of trust validation for {subj} failed.")
         return trusted, revoked, path
 
+    @property
+    def _trust_anchor(self):
+        if self.validation_path is not None:
+            trust_anchor: x509.Certificate = self.validation_path[0]
+            return trust_anchor.subject.human_friendly
+        else:
+            return "No path to trust anchor found."
+
 
 def simple_cms_attribute(attr_type, value):
     """
@@ -414,6 +424,19 @@ def find_cms_attribute(attrs, name):
 #  See RFC 5035.
 
 def as_signing_certificate(cert: x509.Certificate) -> tsp.SigningCertificate:
+    """
+    Format an ASN.1 ``SigningCertificate`` object, where the certificate
+    is identified by its SHA-1 digest.
+
+    .. note::
+        This use of SHA-1 is not cryptographic in nature.
+
+    :param cert:
+        An X.509 certificate.
+    :return:
+        A :class:`tsp.SigningCertificate` object referring to the original
+        certificate.
+    """
     # see RFC 2634, § 5.4.1
     return tsp.SigningCertificate({
         'certs': [
@@ -545,7 +568,19 @@ def _process_pss_params(params: algos.RSASSAPSSParams, digest_algorithm,
     return pss_padding, md
 
 
-def optimal_pss_params(cert: x509.Certificate, digest_algorithm: str):
+def optimal_pss_params(cert: x509.Certificate, digest_algorithm: str) \
+        -> algos.RSASSAPSSParams:
+    """
+    Figure out the optimal RSASSA-PSS parameters for a given certificate.
+    The subject's public key must be an RSA key.
+
+    :param cert:
+        An RSA X.509 certificate.
+    :param digest_algorithm:
+        The digest algorithm to use.
+    :return:
+        RSASSA-PSS parameters.
+    """
 
     digest_algorithm = digest_algorithm.lower()
 
@@ -593,7 +628,7 @@ def load_certs_from_pemder(cert_files):
                 else:  # pragma: nocover
                     logger.debug(
                         f'Skipping PEM block of type {type_name} in '
-                        f'{ca_chain_file}.'
+                        f'CA chain file.'
                     )
         else:
             # no need to unarmor, just try to load it immediately
@@ -707,6 +742,19 @@ def _validate_raw(signature: bytes, signed_data: bytes, cert: x509.Certificate,
         )
 
 
+def extract_message_digest(signer_info: cms.SignerInfo):
+    try:
+        embedded_digest, = find_cms_attribute(
+            signer_info['signed_attrs'], 'message_digest'
+        )
+        return embedded_digest.native
+    except (KeyError, ValueError):
+        raise SignatureValidationError(
+            'Message digest not found in signature, or multiple message '
+            'digest attributes present.'
+        )
+
+
 def validate_sig_integrity(signer_info: cms.SignerInfo,
                            cert: x509.Certificate,
                            expected_content_type: str,
@@ -809,16 +857,7 @@ def validate_sig_integrity(signer_info: cms.SignerInfo,
                 'attributes present.'
             )
 
-        try:
-            embedded_digest, = find_cms_attribute(
-                signed_attrs, 'message_digest'
-            )
-            embedded_digest = embedded_digest.native
-        except (KeyError, ValueError):
-            raise SignatureValidationError(
-                'Message digest not found in signature, or multiple message '
-                'digest attributes present.'
-            )
+        embedded_digest = extract_message_digest(signer_info)
 
         signed_data = signed_attrs.dump()
     try:
