@@ -5,7 +5,7 @@ from collections import namedtuple
 from dataclasses import dataclass, field as data_field
 from datetime import datetime
 from enum import Enum, unique
-from typing import TypeVar, Type, Optional, Union, Iterator
+from typing import TypeVar, Type, Optional, Union, Iterator, IO
 
 from asn1crypto import (
     cms, tsp, ocsp as asn1_ocsp, pdf as asn1_pdf, crl as asn1_crl, x509,
@@ -19,7 +19,8 @@ from pyhanko_certvalidator.path import ValidationPath
 from pyhanko.pdf_utils import generic, misc
 from pyhanko.pdf_utils.generic import pdf_name
 from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
-from pyhanko.pdf_utils.misc import OrderedEnum, get_and_apply
+from pyhanko.pdf_utils.misc import OrderedEnum, get_and_apply, \
+    DEFAULT_CHUNK_SIZE
 from pyhanko.pdf_utils.reader import (
     PdfFileReader, XRefCache, process_data_at_eof,
 )
@@ -37,18 +38,20 @@ from .general import (
     UnacceptableSignerError, KeyUsageConstraints,
     SignatureValidationError,
     validate_sig_integrity, DEFAULT_WEAK_HASH_ALGORITHMS,
-    get_pyca_cryptography_hash,
+    get_pyca_cryptography_hash, extract_message_digest,
 )
 from .timestamps import TimestampSignatureStatus
 
 __all__ = [
-    'SignatureCoverageLevel', 'PdfSignatureStatus',
+    'SignatureCoverageLevel', 'PdfSignatureStatus', 'DocumentTimestampStatus',
+    'StandardCMSSignatureStatus', 'ModificationInfo',
     'EmbeddedPdfSignature', 'DocMDPInfo',
     'RevocationInfoValidationType', 'VRI', 'DocumentSecurityStore',
     'apply_adobe_revocation_info', 'get_timestamp_chain',
     'read_certification_data', 'validate_pdf_ltv_signature',
     'validate_pdf_signature', 'validate_cms_signature',
-    'validate_pdf_timestamp', 'collect_validation_info',
+    'validate_detached_cms', 'validate_pdf_timestamp',
+    'collect_validation_info',
     'add_validation_info',
     'ValidationInfoReadingError', 'SigSeedValueValidationError'
 ]
@@ -77,23 +80,34 @@ class SigSeedValueValidationError(SignatureValidationError):
         super().__init__(failure_message)
 
 
+def _get_signer_predicate(sid: cms.SignerIdentifier):
+    if sid.name == 'issuer_and_serial_number':
+        issuer = sid.chosen['issuer']
+        serial_number = sid.chosen['serial_number'].native
+        return lambda c: c.issuer == issuer and c.serial_number == serial_number
+    elif sid.name == 'subject_key_identifier':
+        # subject key identifier (not all certs have this, but that shouldn't
+        # be an issue in this case)
+        ski = sid.chosen.native
+        logger.warning(
+            "The signature in this file seems to be identified by a subject "
+            "key identifier --- this is legal in CMS, but many PDF viewers and "
+            "SDKs do not support this."
+        )
+        return lambda c: c.key_identifier == ski
+    raise NotImplementedError
+
+
 def partition_certs(certs, signer_info):
     # The 'certificates' entry is defined as a set in PCKS#7.
     # In particular, we cannot make any assumptions about the order.
     # This means that we have to manually dig through the list to find
     # the actual signer
-    iss_sn = signer_info['sid']
-    # TODO Figure out how the subject key identifier thing works
-    if iss_sn.name != 'issuer_and_serial_number':
-        raise NotImplementedError(
-            'Can only look up certificates by issuer and serial number'
-        )
-    issuer = iss_sn.chosen['issuer']
-    serial_number = iss_sn.chosen['serial_number'].native
+    predicate = _get_signer_predicate(signer_info['sid'])
     cert = None
     other_certs = []
     for c in certs:
-        if c.issuer == issuer and c.serial_number == serial_number:
+        if predicate(c):
             cert = c
         else:
             other_certs.append(c)
@@ -107,16 +121,61 @@ def partition_certs(certs, signer_info):
 StatusType = TypeVar('StatusType', bound=SignatureStatus)
 
 
-def _extract_signer_info_and_certs(signed_data: cms.SignedData):
-    certs = [c.parse() for c in signed_data['certificates']]
-
+def _extract_signer_info(signed_data: cms.SignedData) -> cms.SignerInfo:
     try:
         signer_info, = signed_data['signer_infos']
+        return signer_info
     except ValueError:  # pragma: nocover
         raise ValueError(
             'signer_infos should contain exactly one entry'
         )
 
+
+def _extract_self_reported_ts(signer_info: cms.SignerInfo) \
+        -> Optional[datetime]:
+    try:
+        sa = signer_info['signed_attrs']
+        st = find_cms_attribute(sa, 'signing_time')[0]
+        return st.native
+    except KeyError:
+        pass
+
+
+def _extract_tst_data(signer_info, signed=False) -> Optional[cms.SignedData]:
+    try:
+        if signed:
+            sa = signer_info['signed_attrs']
+            tst = find_cms_attribute(sa, 'content_time_stamp')[0]
+        else:
+            ua = signer_info['unsigned_attrs']
+            tst = find_cms_attribute(ua, 'signature_time_stamp_token')[0]
+        tst_signed_data = tst['content']
+        return tst_signed_data
+    except KeyError:
+        pass
+
+
+def _compute_tst_digest(signer_info: cms.SignerInfo) -> Optional[bytes]:
+
+    tst_data = _extract_tst_data(signer_info)
+    if tst_data is None:
+        return None
+
+    eci = tst_data['encap_content_info']
+    mi = eci['content'].parsed['message_imprint']
+    tst_md_algorithm = mi['hash_algorithm']['algorithm'].native
+
+    signature_bytes = signer_info['signature'].native
+    tst_md_spec = get_pyca_cryptography_hash(tst_md_algorithm)
+    md = hashes.Hash(tst_md_spec)
+    md.update(signature_bytes)
+    return md.finalize()
+
+
+def _extract_signer_info_and_certs(signed_data: cms.SignedData):
+    certs = [c.parse() for c in signed_data['certificates']]
+
+    signer_info = _extract_signer_info(signed_data)
     cert, other_certs = partition_certs(certs, signer_info)
 
     return signer_info, cert, other_certs
@@ -144,13 +203,11 @@ def _validate_cms_signature(signed_data: cms.SignedData,
         signer_info['signature_algorithm']
     mechanism = signature_algorithm['algorithm'].native
     md_algorithm = signer_info['digest_algorithm']['algorithm'].native
-    expected_content_type = 'data'
+    eci = signed_data['encap_content_info']
+    expected_content_type = eci['content_type'].native
     if raw_digest is None:
         # this means that there should be encapsulated data
-        eci = signed_data['encap_content_info']
-        expected_content_type = eci['content_type'].native
-
-        raw = eci['content'].parsed.dump()
+        raw = bytes(eci['content'])
         md_spec = get_pyca_cryptography_hash(md_algorithm)
         md = hashes.Hash(md_spec)
         md.update(raw)
@@ -197,7 +254,10 @@ def validate_cms_signature(signed_data: cms.SignedData,
                            key_usage_settings: KeyUsageConstraints = None,
                            encap_data_invalid=False):
     """
-    Validate a detached CMS signature (i.e. a ``SignedData`` object).
+    Validate a CMS signature (i.e. a ``SignedData`` object).
+
+    .. versionchanged:: 0.7.0
+        Now handles both detached and enveloping signatures.
 
     :param signed_data:
         The :class:`.asn1crypto.cms.SignedData` object to validate.
@@ -222,7 +282,7 @@ def validate_cms_signature(signed_data: cms.SignedData,
         This option is considered internal API, the semantics of which may
         change without notice in the future.
     :return:
-        A :class:`SignatureStatus` object (or an instance of a proper subclass)
+        A :class:`.SignatureStatus` object (or an instance of a proper subclass)
     """
     status_kwargs = _validate_cms_signature(
         signed_data, status_cls, raw_digest, validation_context,
@@ -231,6 +291,232 @@ def validate_cms_signature(signed_data: cms.SignedData,
     return status_cls(**status_kwargs)
 
 
+def collect_timing_info(signer_info: cms.SignerInfo,
+                        ts_validation_context: ValidationContext):
+
+    status_kwargs = {}
+
+    # timestamp-related validation
+    signer_reported_dt = _extract_self_reported_ts(signer_info)
+    if signer_reported_dt is not None:
+        status_kwargs['signer_reported_dt'] = signer_reported_dt
+
+    tst_signed_data = _extract_tst_data(signer_info, signed=False)
+    if tst_signed_data is not None:
+        tst_validity_kwargs = _validate_timestamp(
+            tst_signed_data, ts_validation_context,
+            _compute_tst_digest(signer_info),
+        )
+        tst_validity = TimestampSignatureStatus(**tst_validity_kwargs)
+        status_kwargs['timestamp_validity'] = tst_validity
+
+    content_tst_signed_data = _extract_tst_data(signer_info, signed=True)
+    if content_tst_signed_data is not None:
+        content_tst_validity_kwargs = _validate_timestamp(
+            content_tst_signed_data, ts_validation_context,
+            expected_tst_imprint=extract_message_digest(signer_info)
+        )
+        content_tst_validity = TimestampSignatureStatus(
+            **content_tst_validity_kwargs
+        )
+        status_kwargs['content_timestamp_validity'] = content_tst_validity
+
+    return status_kwargs
+
+
+@dataclass(frozen=True)
+class StandardCMSSignatureStatus(SignatureStatus):
+    """
+    Status of a standard "end-entity" CMS signature, potentially with
+    timing information embedded inside.
+    """
+
+    signer_reported_dt: Optional[datetime] = None
+    """
+    Signer-reported signing time, if present in the signature.
+
+    Generally speaking, this timestamp should not be taken as fact.
+    """
+
+    timestamp_validity: Optional[TimestampSignatureStatus] = None
+    """
+    Validation status of the signature timestamp token embedded in this 
+    signature, if present.
+    """
+
+    content_timestamp_validity: Optional[TimestampSignatureStatus] = None
+    """
+    Validation status of the content timestamp token embedded in this 
+    signature, if present.
+    """
+
+    @property
+    def bottom_line(self) -> bool:
+        """
+        Formulates a general judgment on the validity of this signature.
+        This takes into account the cryptographic validity of the signature,
+        the signature's chain of trust and the validity of the timestamp token
+        (if present).
+
+        :return:
+            ``True`` if all constraints are satisfied, ``False`` otherwise.
+        """
+
+        ts = self.timestamp_validity
+        if ts is None:
+            timestamp_ok = True
+        else:
+            timestamp_ok = ts.valid and ts.trusted
+
+        content_ts = self.content_timestamp_validity
+        if content_ts is None:
+            content_timestamp_ok = True
+        else:
+            content_timestamp_ok = content_ts.valid and content_ts.trusted
+        return (
+                self.intact and self.valid and self.trusted and timestamp_ok
+                and content_timestamp_ok
+        )
+
+    def summary_fields(self):
+        yield from super().summary_fields()
+
+        if self.timestamp_validity is not None:
+            yield 'TIMESTAMP_TOKEN<%s>' % (
+                '|'.join(self.timestamp_validity.summary_fields())
+            )
+
+    def pretty_print_details(self):
+        def fmt_section(hdr, body):
+            return '\n'.join(
+                (hdr, '-' * len(hdr), body, '\n')
+            )
+        sections = self.pretty_print_sections()
+        bottom_line = (
+            f"The signature is judged {'' if self.bottom_line else 'IN'}VALID."
+        )
+        sections.append(("Bottom line", bottom_line))
+        return '\n'.join(
+            fmt_section(hdr, body) for hdr, body in sections
+        )
+
+    def pretty_print_sections(self):
+        cert: x509.Certificate = self.signing_cert
+
+        if self.trusted:
+            trust_status = "trusted"
+        elif self.revoked:
+            trust_status = "revoked"
+        else:
+            trust_status = "untrusted"
+        about_signer = (
+            f"Certificate subject: \"{cert.subject.human_friendly}\"\n"
+            f"Certificate SHA1 fingerprint: {cert.sha1.hex()}\n"
+            f"Certificate SHA256 fingerprint: {cert.sha256.hex()}\n"
+            f"Trust anchor: \"{self._trust_anchor}\"\n"
+            f"The signer's certificate is {trust_status}."
+        )
+
+        validity_info = (
+            "The signature is cryptographically "
+            f"{'' if self.intact and self.valid else 'un'}sound."
+        )
+
+        timing_infos = []
+        reported_ts = self.signer_reported_dt
+        if reported_ts is not None:
+            timing_infos.append(
+                f"Signing time as reported by signer: {reported_ts.isoformat()}"
+            )
+
+        tst_status = self.timestamp_validity
+        if tst_status is not None:
+            ts = tst_status.timestamp
+            timing_infos.append(
+                f"Signature timestamp token: {ts.isoformat()}\n"
+                f"The token is guaranteed to be newer than the signature.\n"
+                f"{tst_status.describe_timestamp_trust()}"
+            )
+        content_tst_status = self.timestamp_validity
+        if tst_status is not None:
+            ts = content_tst_status.timestamp
+            timing_infos.append(
+                f"Content timestamp token: {ts.isoformat()}\n"
+                f"The token is guaranteed to be older than the signature.\n"
+                f"{content_tst_status.describe_timestamp_trust()}"
+            )
+        timing_info = (
+            "No available information about the signing time."
+            if not timing_infos else '\n\n'.join(timing_infos)
+        )
+        return [
+            ("Signer info", about_signer), ("Integrity", validity_info),
+            ("Signing time", timing_info),
+        ]
+
+
+def validate_detached_cms(input_data: Union[bytes, IO,
+                                            cms.ContentInfo,
+                                            cms.EncapsulatedContentInfo],
+                          signed_data: cms.SignedData,
+                          signer_validation_context: ValidationContext = None,
+                          ts_validation_context: ValidationContext = None,
+                          key_usage_settings: KeyUsageConstraints = None,
+                          chunk_size=DEFAULT_CHUNK_SIZE,
+                          max_read=None) -> StandardCMSSignatureStatus:
+    """
+    .. versionadded: 0.7.0
+
+    Validate a detached CMS signature.
+
+    :param input_data:
+        The input data to sign. This can be either a :class:`bytes` object,
+        a file-like object or a :class:`cms.ContentInfo` /
+        :class:`cms.EncapsulatedContentInfo` object.
+
+        If a CMS content info object is passed in, the `content` field
+        will be extracted.
+    :param signed_data:
+        The :class:`cms.SignedData` object containing the signature to verify.
+    :param signer_validation_context:
+        Validation context to use to verify the signer certificate's trust.
+    :param ts_validation_context:
+        Validation context to use to verify the TSA certificate's trust, if
+        a timestamp token is present.
+        By default, the same validation context as that of the signer is used.
+    :param key_usage_settings:
+        Key usage parameters for the signer.
+    :param chunk_size:
+        Chunk size to use when consuming input data.
+    :param max_read:
+        Maximal number of bytes to read from the input stream.
+    :return:
+        A description of the signature's status.
+    """
+
+    if ts_validation_context is None:
+        ts_validation_context = signer_validation_context
+    signer_info = _extract_signer_info(signed_data)
+    digest_algorithm = signer_info['digest_algorithm']['algorithm'].native
+    h = hashes.Hash(get_pyca_cryptography_hash(digest_algorithm))
+    if isinstance(input_data, bytes):
+        h.update(input_data)
+    elif isinstance(input_data, (cms.ContentInfo, cms.EncapsulatedContentInfo)):
+        h.update(bytes(input_data['content']))
+    else:
+        temp_buf = bytearray(chunk_size)
+        misc.chunked_digest(temp_buf, input_data, h, max_read=max_read)
+    digest_bytes = h.finalize()
+
+    return validate_cms_signature(
+        signed_data, status_cls=StandardCMSSignatureStatus,
+        raw_digest=digest_bytes,
+        validation_context=signer_validation_context,
+        status_kwargs=collect_timing_info(signer_info, ts_validation_context),
+        key_usage_settings=key_usage_settings
+    )
+
+
 @unique
 class SignatureCoverageLevel(OrderedEnum):
     """
@@ -281,9 +567,9 @@ class ModificationInfo:
 
     * If ``None``, no difference analysis was run.
     * If the difference analysis was successful, this attribute will contain
-      a :class:`DiffResult` object.
+      a :class:`.DiffResult` object.
     * If the difference analysis failed due to unforeseen or suspicious
-      modifications, the :class:`SuspiciousModification` exception thrown
+      modifications, the :class:`.SuspiciousModification` exception thrown
       by the difference policy will be stored in this attribute.
     """
 
@@ -314,30 +600,18 @@ class ModificationInfo:
 
 
 @dataclass(frozen=True)
-class PdfSignatureStatus(ModificationInfo, SignatureStatus):
+class PdfSignatureStatus(ModificationInfo, StandardCMSSignatureStatus):
     """Class to indicate the validation status of a PDF signature."""
 
     docmdp_ok: Optional[bool] = None
     """
-    Indicates whether the signature's :attr:`modification_level` is in line with
-    the document signature policy in force.
+    Indicates whether the signature's
+    :attr:`~.ModificationInfo.modification_level` is in line with the document
+    signature policy in force.
     
     If ``None``, compliance could not be determined.
     """
 
-    signer_reported_dt: Optional[datetime] = None
-    """
-    Signer-reported signing time, if present in the signature.
-    
-    Generally speaking, this timestamp should not be taken as fact.
-    """
-
-    timestamp_validity: Optional[TimestampSignatureStatus] = None
-    """
-    Validation status of the timestamp token embedded in this signature, 
-    if present.
-    """
-
     has_seed_values: bool = False
     """
     Records whether the signature form field has seed values.
@@ -366,9 +640,9 @@ class PdfSignatureStatus(ModificationInfo, SignatureStatus):
         if ts is None:
             timestamp_ok = True
         else:
-            timestamp_ok = ts.valid and ts.trusted
+            timestamp_ok = ts.intact and ts.valid and ts.trusted
         return (
-            self.valid and self.trusted and self.seed_value_ok
+            self.intact and self.valid and self.trusted and self.seed_value_ok
             and (self.docmdp_ok or self.modification_level is None)
             and timestamp_ok
         )
@@ -405,35 +679,9 @@ class PdfSignatureStatus(ModificationInfo, SignatureStatus):
                 yield 'ACCEPTABLE_MODIFICATIONS'
         else:
             yield 'ILLEGAL_MODIFICATIONS'
-        if self.timestamp_validity is not None:
-            yield 'TIMESTAMP_TOKEN<%s>' % (
-                '|'.join(self.timestamp_validity.summary_fields())
-            )
-
-    def pretty_print_details(self):
-        cert: x509.Certificate = self.signing_cert
-
-        def _trust_anchor(status: SignatureStatus):
-            if status.validation_path is not None:
-                trust_anchor: x509.Certificate = status.validation_path[0]
-                return trust_anchor.subject.human_friendly
-            else:
-                return "No path to trust anchor found."
-
-        if self.trusted:
-            trust_status = "trusted"
-        elif self.revoked:
-            trust_status = "revoked"
-        else:
-            trust_status = "untrusted"
-        about_signer = (
-            f"Certificate subject: \"{cert.subject.human_friendly}\"\n"
-            f"Certificate SHA1 fingerprint: {cert.sha1.hex()}\n"
-            f"Certificate SHA256 fingerprint: {cert.sha256.hex()}\n"
-            f"Trust anchor: \"{_trust_anchor(self)}\"\n"
-            f"The signer's certificate is {trust_status}."
-        )
 
+    def pretty_print_sections(self):
+        sections = super().pretty_print_sections()
         if self.coverage == SignatureCoverageLevel.ENTIRE_FILE:
             modification_str = "The signature covers the entire file."
         else:
@@ -457,53 +705,7 @@ class PdfSignatureStatus(ModificationInfo, SignatureStatus):
             else:
                 modification_str = "Incremental update analysis was skipped"
 
-        validity_info = (
-            "The signature is cryptographically "
-            f"{'' if self.intact and self.valid else 'un'}sound.\n"
-            f"{modification_str}"
-        )
-
-        ts = self.signer_reported_dt
-        tst_status = self.timestamp_validity
-        about_tsa = ''
-        if tst_status is not None:
-            ts = tst_status.timestamp
-            tsa = tst_status.signing_cert
-
-            about_tsa = (
-                "The signing time is guaranteed by a time stamping authority.\n"
-                f"TSA certificate subject: \"{tsa.subject.human_friendly}\"\n"
-                f"TSA certificate SHA1 fingerprint: {tsa.sha1.hex()}\n"
-                f"TSA certificate SHA256 fingerprint: {tsa.sha256.hex()}\n"
-                f"TSA cert trust anchor: \"{_trust_anchor(tst_status)}\"\n"
-                "The TSA certificate is "
-                f"{'' if tst_status.trusted else 'un'}trusted."
-            )
-        elif ts is not None:
-            about_tsa = "The signing time is self-reported by the signer."
-
-        if ts is not None:
-            signing_time_str = ts.isoformat()
-        else:
-            signing_time_str = "unknown"
-
-        timing_info = (
-            f"Signing time: {signing_time_str}\n{about_tsa}"
-        )
-
-        def fmt_section(hdr, body):
-            return '\n'.join(
-                (hdr, '-' * len(hdr), body, '\n')
-            )
-
-        bottom_line = (
-            f"The signature is judged {'' if self.bottom_line else 'IN'}VALID."
-        )
-
-        sections = [
-            ("Signer info", about_signer), ("Integrity", validity_info),
-            ("Signing time", timing_info),
-        ]
+        sections.append(("Modifications", modification_str))
         if self.has_seed_values:
             if self.seed_value_ok:
                 sv_info = "There were no SV issues detected for this signature."
@@ -515,10 +717,7 @@ class PdfSignatureStatus(ModificationInfo, SignatureStatus):
                 )
             sections.append(("Seed value constraints", sv_info))
 
-        sections.append(("Bottom line", bottom_line))
-        return '\n'.join(
-            fmt_section(hdr, body) for hdr, body in sections
-        )
+        return sections
 
 
 @dataclass(frozen=True)
@@ -682,7 +881,6 @@ class EmbeddedPdfSignature:
         """
         return self.fq_name
 
-    # TODO also parse the signature object's /M entry
     @property
     def self_reported_timestamp(self) -> Optional[datetime]:
         """
@@ -690,12 +888,9 @@ class EmbeddedPdfSignature:
             The signing time as reported by the signer, if embedded in the
             signature's signed attributes.
         """
-        try:
-            sa = self.signer_info['signed_attrs']
-            st = find_cms_attribute(sa, 'signed_time')[0]
-            return st.native
-        except KeyError:
-            pass
+        ts = _extract_self_reported_ts(self.signer_info)
+        if ts is not None:
+            return ts
 
         try:
             st_as_pdf_date = self.sig_object['/M']
@@ -710,13 +905,7 @@ class EmbeddedPdfSignature:
             The signed data component of the timestamp token embedded in this
             signature, if present.
         """
-        try:
-            ua = self.signer_info['unsigned_attrs']
-            tst = find_cms_attribute(ua, 'signature_time_stamp_token')[0]
-            tst_signed_data = tst['content']
-            return tst_signed_data
-        except KeyError:
-            pass
+        return _extract_tst_data(self.signer_info)
 
     def compute_integrity_info(self, diff_policy=None, skip_diff=False):
         """
@@ -725,7 +914,8 @@ class EmbeddedPdfSignature:
         :param diff_policy:
             Policy to evaluate potential incremental updates that were appended
             to the signed revision of the document.
-            Defaults to :attr:`.DEFAULT_DIFF_POLICY`.
+            Defaults to
+            :const:`~pyhanko.sign.diff_analysis.DEFAULT_DIFF_POLICY`.
         :param skip_diff:
             If ``True``, skip the difference analysis step entirely.
         """
@@ -744,7 +934,7 @@ class EmbeddedPdfSignature:
     def summarise_integrity_info(self) -> dict:
         """
         Compile the integrity information for this signature into a dictionary
-        that can later be passed to :class:`PdfSignatureStatus` as kwargs.
+        that can later be passed to :class:`.PdfSignatureStatus` as kwargs.
 
         This method is only available after calling
         :meth:`.EmbeddedPdfSignature.compute_integrity_info`.
@@ -896,21 +1086,9 @@ class EmbeddedPdfSignature:
 
         if self.tst_signature_digest is not None:
             return self.tst_signature_digest
-        # for timestamp validation: compute the digest of the signature
-        #  (as embedded in the CMS object)
-        tst_data = self.attached_timestamp_data
-        if tst_data is None:
-            return None
-
-        eci = tst_data['encap_content_info']
-        mi = eci['content'].parsed['message_imprint']
-        tst_md_algorithm = mi['hash_algorithm']['algorithm'].native
-
-        signature_bytes = self.signer_info['signature'].native
-        tst_md_spec = get_pyca_cryptography_hash(tst_md_algorithm)
-        md = hashes.Hash(tst_md_spec)
-        md.update(signature_bytes)
-        self.tst_signature_digest = digest = md.finalize()
+        self.tst_signature_digest = digest = _compute_tst_digest(
+            self.signer_info
+        )
         return digest
 
     def evaluate_signature_coverage(self) -> SignatureCoverageLevel:
@@ -1210,7 +1388,8 @@ def validate_pdf_signature(embedded_sig: EmbeddedPdfSignature,
     :param diff_policy:
         Policy to evaluate potential incremental updates that were appended
         to the signed revision of the document.
-        Defaults to :attr:`.DEFAULT_DIFF_POLICY`.
+        Defaults to
+        :const:`~pyhanko.sign.diff_analysis.DEFAULT_DIFF_POLICY`.
     :param key_usage_settings:
         A :class:`.KeyUsageConstraints` object specifying which key usage
         extensions must or must not be present in the signer's certificate.
@@ -1240,25 +1419,14 @@ def validate_pdf_signature(embedded_sig: EmbeddedPdfSignature,
     )
     status_kwargs = embedded_sig.summarise_integrity_info()
 
-    # try to find an embedded timestamp
-    signer_reported_dt = embedded_sig.self_reported_timestamp
-    if signer_reported_dt is not None:
-        status_kwargs['signer_reported_dt'] = signer_reported_dt
-
-    # if we managed to find an (externally) signed timestamp,
-    # we now proceed to validate it
-    tst_signed_data = embedded_sig.attached_timestamp_data
-    # TODO compare value of embedded timestamp token with the timestamp
-    #  attribute if both are present
-    tst_validity: Optional[SignatureStatus] = None
-    if tst_signed_data is not None:
-        assert embedded_sig.tst_signature_digest is not None
-        tst_validity_kwargs = _validate_timestamp(
-            tst_signed_data, ts_validation_context,
-            embedded_sig.tst_signature_digest
-        )
-        tst_validity = TimestampSignatureStatus(**tst_validity_kwargs)
-        status_kwargs['timestamp_validity'] = tst_validity
+    status_kwargs.update(
+        collect_timing_info(embedded_sig.signer_info, ts_validation_context)
+    )
+    if 'signer_reported_dt' not in status_kwargs:
+        # maybe the PDF signature dictionary declares /M
+        signer_reported_dt = embedded_sig.self_reported_timestamp
+        if signer_reported_dt is not None:
+            status_kwargs['signer_reported_dt'] = signer_reported_dt
 
     status_kwargs = _validate_cms_signature(
         embedded_sig.signed_data, status_cls=PdfSignatureStatus,
@@ -1266,6 +1434,7 @@ def validate_pdf_signature(embedded_sig: EmbeddedPdfSignature,
         validation_context=signer_validation_context,
         status_kwargs=status_kwargs, key_usage_settings=key_usage_settings
     )
+    tst_validity = status_kwargs.get('timestamp_validity', None)
     timestamp_found = (
         tst_validity is not None
         and tst_validity.valid and tst_validity.trusted
@@ -1288,7 +1457,8 @@ def validate_pdf_timestamp(embedded_sig: EmbeddedPdfSignature,
     :param diff_policy:
         Policy to evaluate potential incremental updates that were appended
         to the signed revision of the document.
-        Defaults to :attr:`.DEFAULT_DIFF_POLICY`.
+        Defaults to
+        :const:`~pyhanko.sign.diff_analysis.DEFAULT_DIFF_POLICY`.
     :param skip_diff:
         If ``True``, skip the difference analysis step entirely.
     :return:
@@ -1482,8 +1652,9 @@ def validate_pdf_ltv_signature(embedded_sig: EmbeddedPdfSignature,
     :param validation_type:
         Validation profile to use.
     :param validation_context_kwargs:
-        Keyword args to instantiate :class:`.pyhanko_certvalidator.ValidationContext`
-        objects needed over the course of the validation.
+        Keyword args to instantiate
+        :class:`.pyhanko_certvalidator.ValidationContext` objects needed over
+        the course of the validation.
     :param bootstrap_validation_context:
         Validation context used to validate the current timestamp.
     :param force_revinfo:
@@ -1492,7 +1663,8 @@ def validate_pdf_ltv_signature(embedded_sig: EmbeddedPdfSignature,
     :param diff_policy:
         Policy to evaluate potential incremental updates that were appended
         to the signed revision of the document.
-        Defaults to :attr:`.DEFAULT_DIFF_POLICY`.
+        Defaults to
+        :const:`~pyhanko.sign.diff_analysis.DEFAULT_DIFF_POLICY`.
     :param key_usage_settings:
         A :class:`.KeyUsageConstraints` object specifying which key usage
         extensions must or must not be present in the signer's certificate.
@@ -1848,7 +2020,8 @@ def collect_validation_info(embedded_sig: EmbeddedPdfSignature,
 def add_validation_info(embedded_sig: EmbeddedPdfSignature,
                         validation_context: ValidationContext,
                         skip_timestamp=False, add_vri_entry=True,
-                        in_place=False, output=None, chunk_size=4096):
+                        in_place=False, output=None,
+                        chunk_size=DEFAULT_CHUNK_SIZE):
     """
     Add validation info (CRLs, OCSP responses, extra certificates) for a
     signature to the DSS of a document in an incremental update.
@@ -2047,10 +2220,9 @@ class DocumentSecurityStore:
             A PDF object representing this DSS.
         """
         pdf_dict = self.backing_pdf_object
-        pdf_dict.update({
-            pdf_name('/VRI'): generic.DictionaryObject(self.vri_entries),
-            pdf_name('/Certs'): generic.ArrayObject(list(self.certs.values())),
-        })
+        pdf_dict['/Certs'] = generic.ArrayObject(list(self.certs.values()))
+        if self.vri_entries:
+            pdf_dict['/VRI'] = generic.DictionaryObject(self.vri_entries)
 
         if self.ocsps:
             pdf_dict[pdf_name('/OCSPs')] = generic.ArrayObject(self.ocsps)
